/** \file dm9000.c
\brief generic Davicom DM9000 (E, A, B) driver with some lwIP flavour
\note Based on the Linux kernel driver, see dm9000.c there for copyright and license info

The driver supports the original DM9000E, and now the two newer
devices, DM9000A and DM9000B. Define which one you are using

\todo Link state
\todo MAC Multicast support (external hash-table tool)
\todo resiliency (reset device on DM9000_PKT_ERR in rx function)
*/

#include <stdint.h>
#include <stdio.h>

//#include "arch/cc.h"	// lwIP-port, compiler/processor dependencies
#include "dm9000.h"
#include "dm9000x.h"

#include "dm9000x_macros.h"

#ifdef DM9000_POLL_CONTROLLER
	#define dm9000_disableint(a)
	#define dm9000_enableint(b)
#else
	#include "sys_intsupport.h"
	#define dm9000_disableint(a)	sys_disableint(a)
	#define dm9000_enableint(b)	sys_enableint(b)
#endif

static void dm9000_reset(idtype_chip_address_t db)
{
	/* RESET device */
	idtype_writeidx(db,DM9000_NCR);
//	udelay(200);
	idtype_writedata(db, NCR_RST);
//	udelay(200);
}

/*
 *  Set DM9000 mac and multicast address table
 */
static void dm9000_hash_table(idtype_chip_address_t db, uint8_t *mac_addr, uint16_t *hash_tbl)
{
int i, oft;
uint16_t hash_table[4];
uint8_t rcr = RCR_DIS_LONG | RCR_DIS_CRC | RCR_RXEN;

	for (i = 0, oft = DM9000_PAR; i < 6; i++, oft++)
		idtype_writereg(db, oft, mac_addr[i]);

	/* Clear/copy Hash Table */
	for (i = 0; i < 4; i++){
		if(hash_tbl)
			hash_table[i] = hash_tbl[i];
		else
			hash_table[i] = 0x0;
	}

	/* broadcast address */
	hash_table[3] |= 0x8000;

#ifdef DM9000_PROMISC
		rcr |= RCR_PRMSC;
#endif
#ifdef DM9000_ALLMULTI
		rcr |= RCR_ALL;
#endif

	/* Write the hash table to MAC MD table */
	for (i = 0, oft = DM9000_MAR; i < 4; i++) {
		idtype_writereg(db, oft++, hash_table[i]);
		idtype_writereg(db, oft++, hash_table[i] >> 8);
	}
	idtype_writereg(db, DM9000_RCR, rcr);
}

/*
 * Initialize dm9000 chip
 */
static void
dm9000_init_dm9000(idtype_chip_address_t db, uint8_t *mac_addr, uint16_t *hash_table)
{

#ifdef DM9000_CSUMOFFLOAD
	/* Checksum mode: generate on tx, check and discard on rx */
	idtype_writereg(db, DM9000_TCCR, TCCR_IP | TCCR_TCP	| TCCR_UDP);
	idtype_writereg(db, DM9000_RCSR, RCSR_DISCARD);
#endif

	/* GPIO0 on pre-activate PHY */
	idtype_writereg(db, DM9000_GPR, 0);	/* REG_1F bit0 activate phyxcer */
	idtype_writereg(db, DM9000_GPCR, GPCR_GEP_CNTL);	/* Let GPIO0 output */
	idtype_writereg(db, DM9000_GPR, 0);	/* Enable PHY */

	idtype_writereg(db, DM9000_NCR, DM9000_NCRMASK);

	/* Program operating register */
	idtype_writereg(db, DM9000_TCR, 0);	        /* TX Polling clear */
	idtype_writereg(db, DM9000_BPTR, 0x3f);	/* Less 3Kb, 200us */
	idtype_writereg(db, DM9000_FCR, 0xff);	/* Flow Control */
	idtype_writereg(db, DM9000_SMCR, 0);        /* Special Mode */
	/* clear TX status */
	idtype_writereg(db, DM9000_NSR, NSR_WAKEST | NSR_TX2END | NSR_TX1END);
	idtype_writereg(db, DM9000_ISR, ISR_CLR_STATUS); /* Clear interrupt status */

	/* Set address filter table */
	dm9000_hash_table(db, mac_addr, hash_table);

#ifdef DM9000_CSUMOFFLOAD
/* Current lwIP implementation (1.4.1) is all or nothing,
	so checksum offload will be set here */
			idtype_writereg(db, DM9000_TCCR, TCCR_IP | TCCR_UDP | TCCR_TCP);
#endif

	/* Enable TX/RX interrupt mask */
	idtype_writereg(db, DM9000_IMR, DM9000_INTMASK);
}

static void
dm9000_send_packet(idtype_chip_address_t db, uint16_t pkt_len)
{
	idtype_writereg(db, DM9000_TXPLL, pkt_len);
	idtype_writereg(db, DM9000_TXPLH, pkt_len >> 8);

	idtype_writereg(db, DM9000_TCR, TCR_TXREQ);	// Cleared after TX completes
}

/* Split functions for lwIP access, as frames may come in a group of pbufs
we might need to do copy data from many pbufs for a single frame.
Do call them in a tight loop, since interrupts are disabled from successful
return of start_xmit() to end of end_xmit()
 */
int dm9000_start_xmit(struct idtype_chip_internals *i)
{
idtype_chip_address_t db = i->addr;
volatile struct DM9000_internals *state = i->internals;

	/* If the buffer is filled, we must choose to either:
	 - wait for a fixed time (worst case is ~4ms for 10Mbps)
	 - loop on tx_pkt_cnt (it should be decremented on Tx int)
	 - report the situation to the caller and let him decide */
	if(state->tx_pkt_cnt > 1)
		return 0;

	/* Move data to DM9000 TX RAM */
	idtype_writeidx(db, DM9000_MWCMD);
	idtype_startwrite(db);

	return 1;
}

void dm9000_xmit(struct idtype_chip_internals *i, void *data, uint16_t len)
{
idtype_chip_address_t db = i->addr;

	idtype_writeblock(db, data, len);
}

void dm9000_end_xmit(struct idtype_chip_internals *i, uint16_t total_len)
{
idtype_chip_address_t db = i->addr;
volatile struct DM9000_internals *state = i->internals;

	idtype_endwrite(db);
	dm9000_disableint(i->intnum);

	state->tx_pkt_cnt++;
	/* TX control: First packet immediately sent, second packet queued */
	if (state->tx_pkt_cnt == 1) {
		dm9000_send_packet(db, total_len);
	} else {
		/* Second packet */
		state->queue_pkt_len = total_len;
	}

	dm9000_enableint(i->intnum);
}

/// Aggregated function for generic external access, write a whole frame at once
int dm9000_do_xmit(struct idtype_chip_internals *i, void *data, uint16_t len)
{
	if(dm9000_start_xmit(i)){
		dm9000_xmit(i, data, len);
		dm9000_end_xmit(i, len);
		return 1;
	}
	return 0;
}

/**
\brief Process last received frame
\retval 0	no frame available
\returns	frame length (without CRC)
*/
int dm9000_start_rx(struct idtype_chip_internals *i)
{
idtype_chip_address_t db = i->addr;
volatile struct DM9000_internals *state = i->internals;
__ALIGNto16 struct dm9000_rxhdr rxhdr;
uint8_t rxbyte;
int GoodPacket;
int RxLen;

	if(state->rx_pkt_len)
		return state->rx_pkt_len;		// outstanding frame surviving low-memory condition
	if(state->rx_pkt_cnt == 0)			// no outstanding data (signalled on int)
		return 0;

	do {
		dm9000_disableint(i->intnum);

		idtype_startread(db);
		/* Check packet ready or not */
		idtype_readreg(db, DM9000_MRCMDX);	/* Dummy read */
	
		/* Get most updated data */
		rxbyte = idtype_readdata(db);
	
		/* Status check: this byte must be 0 or 1 */
		if (rxbyte & DM9000_PKT_ERR) {
			// status check fail
			//idtype_writereg(db, DM9000_RCR, 0x00);	/* Stop Device */
			//idtype_writereg(db, DM9000_ISR, IMR_PAR);	/* Stop INT request */
			/** \warning Need to reset the device, this is not a normal state
				*/
			break;
		}
	
		if (!(rxbyte & DM9000_PKT_RDY)){
			state->rx_pkt_cnt=0;
			break;
		}
	
		dm9000_enableint(i->intnum);
		
		/* A packet ready now  & Get status/length */
		GoodPacket = 1;
		idtype_writeidx(db,DM9000_MRCMD);
	
		idtype_readblock(db, &rxhdr, sizeof(rxhdr));
	
	#if (BYTE_ORDER == LITTLE_ENDIAN)
		RxLen = *(uint16_t *)&rxhdr.RxLen[0];
	#else
		RxLen = rxhdr.RxLen[0] + (rxhdr.RxLen[1]<<8);
	#endif
	
		/* Packet Length check (runts and jabbers) */
		if ((RxLen < 0x40) || (RxLen > DM9000_PKT_MAX))
			GoodPacket = 0;
	
		/* rxhdr.RxStatus is identical to RSR register. */
		if (rxhdr.RxStatus & (RSR_FOE | RSR_CE | RSR_AE |
				      RSR_PLE | RSR_RWTO | RSR_LCS | RSR_RF)) {
			GoodPacket = 0;
		}
		if (GoodPacket){
			return RxLen-4;		// for lwIP, need to drop CRC
		} else {
			/* need to dump the packet's data */
			idtype_dumpblock(db, RxLen);
		}
	} while (!GoodPacket);	// check next packet if this one was dropped
	dm9000_enableint(i->intnum);
	return 0;
}

void dm9000_rx(struct idtype_chip_internals *i, void *rdptr, uint16_t RxLen)
{
idtype_chip_address_t db = i->addr;

	/* Read received packet from RX SRAM */
	idtype_readblock(db, rdptr, RxLen);
}

void dm9000_end_rx(struct idtype_chip_internals *i)
{
idtype_chip_address_t db = i->addr;
volatile struct DM9000_internals *state = i->internals;

	idtype_dumpblock(db, 4);	// for lwIP, need to read the CRC from the chip
	state->rx_pkt_len = 0;		// remove possible outstanding frame indication
}

///	decide what to do when the stack has no memory to hold an outstanding frame
void dm9000_rx_nomem(struct idtype_chip_internals *i, uint16_t RxLen)
{
volatile struct DM9000_internals *state = i->internals;

	state->rx_pkt_len = RxLen;	// signal there is an outstanding frame in chip memory, remember its length
}

int dm9000_do_rx(struct idtype_chip_internals *i, void *rdptr)
{
int len;

	if((len=dm9000_start_rx(i))!=0){
		dm9000_rx(i, rdptr, len);
		dm9000_end_rx(i);
		return len;
	}
	return 0;
}

/*
 * DM9000 interrupt handler
 * receive the packet to upper layer, free the transmitted packet
 */

static void dm9000_tx_done(struct idtype_chip_internals *i)
{
idtype_chip_address_t db = i->addr;
volatile struct DM9000_internals *state = i->internals;
int tx_status = idtype_readreg(db, DM9000_NSR);	/* Got TX status */

	if (tx_status & (NSR_TX2END | NSR_TX1END)) {
		/* One packet sent complete */
		state->tx_pkt_cnt--;

		/* Queue packet check & send */
		if (state->tx_pkt_cnt > 0)
			dm9000_send_packet(db, state->queue_pkt_len);
	}
}

void dm9000_interrupt(struct idtype_chip_internals *i)
{
idtype_chip_address_t db = i->addr;
volatile struct DM9000_internals *state = i->internals;
int int_status;
uint8_t reg_save;
extern uint8_t rdptr[];
	
	/** A real interrupt coming 
	\warning we assume an interrupt will not re-fire itself on an embedded hardware.
	If that is not your case, take care of it before calling us
	*/

	/* Save previous register address */
	reg_save = idtype_readidx(db);

	/* Disable all interrupts */
	idtype_writereg(db, DM9000_IMR, IMR_PAR);

	/* Got DM9000 interrupt status */
	int_status = idtype_readreg(db, DM9000_ISR);	/* Got ISR */
	idtype_writereg(db, DM9000_ISR, int_status);	/* Clear ISR status */

	/* There won't be another rx interrupt until there is new received data,
	even though we might already have plenty of data in the chip. */
	if (int_status & ISR_PRS)
		state->rx_pkt_cnt++;


	/* Transmit Interrupt check */
	if (int_status & ISR_PTS)
		dm9000_tx_done(i);

#ifdef	DM9000_LNKCHNGINTERRUPT
	if (int_status & ISR_LNKCHNG)
		i->lsisrcback(i->lsicparam);
#endif

	/* Re-enable interrupt mask */
	idtype_writereg(db, DM9000_IMR, DM9000_INTMASK);

	/* Restore previous register address */
	idtype_writeidx(db, reg_save);
}

#ifdef DM9000_USEWOLINTERRUPT
#error UNSUPPORTED DM9000_USEWOLINTERRUPT
#endif

#ifdef DM9000_POLL_CONTROLLER
void dm9000_poll_controller(struct idtype_chip_internals *i)
{
	dm9000_interrupt(i);
}
#endif

/*
 *  "Open" the interface.
 */
void dm9000_init(struct idtype_chip_internals *i, uint8_t *mac_addr, uint16_t *mcast)
{
idtype_chip_address_t db = i->addr;
volatile struct DM9000_internals *state = i->internals;

	/* Initialize DM9000 chip */
	dm9000_reset(db);
	dm9000_init_dm9000(db, mac_addr, mcast);

	/* Init Driver variables */
	state->tx_pkt_cnt = 0;
	state->queue_pkt_len = 0;
	state->rx_pkt_len = 0;
	state->rx_pkt_cnt = 0;
}

#ifdef DM9000_SUPPORTPHYACCESS
#error UNSUPPORTED DM9000_SUPPORTPHYACCESS
#endif
