/*
idtype:
INDEX/DATA, chips with a pin to select an index register from a data register,
the index register holds the register address and all data accesses are on the
data register. E.g.: DM9000x

16-bit transfers use static variables to allow multiple blocks per packet; their
use is limited to a single chip or single-threaded access with no interleaved
block transfer calls, which is true for lwIP using drivers based on this
structure (e.g.: dm9000)
*/

#ifndef __IDTYPE_H
#define __IDTYPE_H

#include <stdint.h>
#include <stdio.h>

#include "HT32F1655_56.h"

/*
Some bus controllers use byte/half-word addresses depending on bus configuration
That is:
	on an 8-bit bus, external A0 is internal A0
	on a 16-bit bus, external A0 is internal A1

Define IDTYPE_16BUS_BYTEADDRESS if your 16-bit bus' A0 addresses bytes
*/
#ifdef IDTYPE_USE8BUS 
	#error "UNSUPPORTED HARDWARE CONFIGURATION"
#else
#ifdef IDTYPE_USE16BUS 
	#ifdef IDTYPE_16BUS_BYTEADDRESS
		struct idtype_chip_bus {
			uint8_t	idx;
			uint8_t data;
		};
	#else
		struct idtype_chip_bus {
			uint16_t idx;
			uint16_t data;
		};
	#endif
#endif
#endif

typedef union {
	volatile struct idtype_chip_bus *busaddr;
	unsigned int ioidx;
} idtype_chip_address_t;

struct idtype_chip_internals {
	idtype_chip_address_t addr;
	volatile void *internals;
	void (*lsisrcback)();
	void *lsicparam;
	int intnum;
};

#if defined(IDTYPE_USE8BUS) || defined(IDTYPE_USE16BUS)
	#define _readidx(chip)	(chip.busaddr->idx)
	#define _writeidx(chip, reg)	(chip.busaddr->idx=(reg))
	#define _readdata(chip)		(chip.busaddr->data)
	#define _writedata(chip,dat)	(chip.busaddr->data=(dat))
#else
/* if the chip is bit banged on an I/O port, "chip" will lookup related I/Os from
a table, indexed by chip.ioidx */
#endif

#ifdef IDTYPE_USE16BUS 
	void idtype_startread(idtype_chip_address_t chip);
	void idtype_startwrite(idtype_chip_address_t chip);
	void idtype_endwrite(idtype_chip_address_t chip);
#else
	#define idtype_startread(chip)
	#define idtype_startwrite(chip)
	#define idtype_endwrite(chip)
#endif

static __INLINE int idtype_readreg(idtype_chip_address_t chip, int reg)
{
  _writeidx(chip, reg);
  return _readdata(chip);
}

static __INLINE void idtype_writereg(idtype_chip_address_t chip, int reg, int data)
{
  _writeidx(chip, reg);
  _writedata(chip, data);
}

static __INLINE int idtype_readidx(idtype_chip_address_t chip)
{
  return _readidx(chip);
}
static __INLINE void idtype_writeidx(idtype_chip_address_t chip, int reg)
{
  _writeidx(chip,reg);
}

static __INLINE int idtype_readdata(idtype_chip_address_t chip)
{
  return _readdata(chip);
}
static __INLINE void idtype_writedata(idtype_chip_address_t chip, int data)
{
  _writedata(chip,data);
}

void idtype_readblock(idtype_chip_address_t chip, void *data, uint16_t len);

static __INLINE void idtype_dumpblock(idtype_chip_address_t chip, uint16_t len)
{
#ifdef IDTYPE_USE16BUS
register int mylen = (len+1)>>1;
	
	while(mylen--)
		_readdata(chip);
#endif
}

void idtype_writeblock(idtype_chip_address_t chip, void *data, uint16_t len);

#endif
